import can
import struct

def decode_cell_voltage(cell_number, message):
    print(f'For Cell{cell_number}:')
    if message.dlc >= 3:
        voltage_bytes = message.data[0:4]
        voltage_hex = ''.join(f'{byte:02X}' for byte in reversed(voltage_bytes))
        voltage_int = int(voltage_hex, 16)
        formatted_voltage_hex = f'{voltage_int:08X}'
        voltage_data = round(struct.unpack('>f', bytes.fromhex(formatted_voltage_hex))[0], 4)
        print(f'  Voltage: {voltage_data} Volt')
        print(' ')
    else:
        print(f'  Error: Insufficient data for voltage decoding. DLC is less than 4.')


def decode_SoC(message):
    print(f'For SoC:')
    if message.dlc >= 3:
        SoC_bytes = message.data[0:4]
        SoC_hex = ''.join(f'{byte:02X}' for byte in reversed(SoC_bytes))
        SoC_int = int(SoC_hex, 16)
        formatted_SoC_hex = f'{SoC_int:08X}'
        SoC_data = round(struct.unpack('>f', bytes.fromhex(formatted_SoC_hex))[0], 4)
        print(f'  SoC: %{SoC_data}')
        print(' ')
    else:
        print(f'  Error: Insufficient data for SoC decoding. DLC is less than 4.')

def decode_curLim(message):
    print(f'For current limit dependin on SoC value:')
    if message.dlc >= 3:
        SoC_bytes = message.data[0:4]
        SoC_hex = ''.join(f'{byte:02X}' for byte in reversed(SoC_bytes))
        SoC_int = int(SoC_hex, 16)
        formatted_SoC_hex = f'{SoC_int:08X}'
        SoC_data = round(struct.unpack('>f', bytes.fromhex(formatted_SoC_hex))[0], 4)
        print(f'  Curent: {SoC_data}Amp')
        print(' ')
    else:
        print(f'  Error: Insufficient data for SoC decoding. DLC is less than 4.')


def decode_temperature(message):
    print('For Temperature:')
    if message.dlc >= 3:
        temperature_bytes = message.data[0:4]
        temperature_hex = ''.join(f'{byte:02X}' for byte in reversed(temperature_bytes))
        temperature_int = int(temperature_hex, 16)
        formatted_temperature_hex = f'{temperature_int:08X}'
        temperature_data = round(struct.unpack('>f', bytes.fromhex(formatted_temperature_hex))[0], 1)
        print(f'  Temperature: {temperature_data} Celsius')
        print(' ')
    else:
        print(f'  Error: Insufficient data for temperature decoding. DLC is less than 4.')

def decode_can_message(message):

    if message.is_remote_frame:
        #print("Remote request message received. Ignoring.")
        return
        
    pgn = (message.arbitration_id >> 8) & 0xFFFF
    print(f'Received CAN Message:')
    print(f'  Arbitration ID: 0x{message.arbitration_id:X}')
    print(f'  Data Length Code: {message.dlc}')

    if pgn in range(1, 250):
        decode_cell_voltage(pgn, message)

    elif pgn in range(4097, 4099):
        decode_temperature(message)
        
    elif pgn == 64772:
        decode_SoC(message)
        
    elif pgn == 8196:
        decode_curLim(message)

    elif pgn == 65281:
        print(f'Request Sent from a CAN device:')

    else:
        print(f'Unknown PGN: {pgn}')

def receive_can_messages():
    channel = 'can0'  # Adjust this to your actual CAN channel
    bus = can.interface.Bus(channel=channel, bustype='socketcan')

    try:
        while True:
            message = bus.recv()
            decode_can_message(message)

    except KeyboardInterrupt:
        print("KeyboardInterrupt: Stopping CAN receiver.")

    finally:
        bus.shutdown()

if __name__ == "__main__":
    receive_can_messages()

